import sys
import os
import copy
import random
from multiprocessing import Pool
from ete2 import Tree
from ete2 import PhyloTree
from ete2.treeview.main import TreeStyle, NodeStyle, TreeStyle
from ete2 import faces, layouts

sys.path.append("../../")

from soft.utils.camber_params_utils import *
from ressoft.utils.file_utils import ensure_dir
from ressoft.utils.stats_utils import *
from ressoft.utils.generator import choose
from ressoft.utils.pred_utils import *
from ressoft.utils.input_utils import *
from ressoft.utils.names_utils import readClusters, readClusterNames


def calcResistantNodeLeaves(phylo_tree):
    mutated_nodes = set([])
    mutated_node_leaves = {}
    queue = deque([])
    for node in phylo_tree.nodes.values():
        node.visited = 0

    for i in phylo_tree.leaf_ids:
        node = phylo_tree.nodes[i]
        queue.append(node.parent)
                
    while queue:
        node = queue.popleft()
        if node == None:
            continue
        node.visited += 1
        if node.visited < len(node.children):
            continue
        i = node.node_id
        
        children_list = list(node.children)
        node_mut = node.complete_res_state

        for j in range(len(children_list)):    
            node_j = children_list[j]  
                        
            node_j_mut = node_j.complete_res_state
            
            if node_j_mut == 'R' and node_mut == 'S':
                mutated_nodes.add(node_j)
        queue.append(node.parent)
    
    for node in mutated_nodes:
        mutated_node_leaves[node.node_id] = node.getLeaves()
    
    return mutated_node_leaves

def calcMutatedNodeLeaves(phylo_tree):
    mutated_nodes = set([])
    mutated_node_leaves = {}
    queue = deque([])
    for node in phylo_tree.nodes.values():
        node.visited = 0

    for i in phylo_tree.leaf_ids:
        node = phylo_tree.nodes[i]
        queue.append(node.parent)
                
    while queue:
        node = queue.popleft()
        if node == None:
            continue
        node.visited += 1
        if node.visited < len(node.children):
            continue
        i = node.node_id
        
        children_list = list(node.children)
        node_mut = node.complete_mut_state

        for j in range(len(children_list)):    
            node_j = children_list[j]  
                        
            node_j_mut = node_j.complete_mut_state
            
            if node_j_mut == '1' and node_mut == '0':
                mutated_nodes.add(node_j)
        queue.append(node.parent)
    
    for node in mutated_nodes:
        mutated_node_leaves[node.node_id] = node.getLeaves()
    
    return mutated_node_leaves


def scoreMutations(drug_res_profile, drug_name, drug_targets, drug_res_genes, tree, all_mutations, gene_presence, strains, output_format_fh=None, parameters=None):
    dataset = parameters["D"] 
    details = parameters["DETAILS"]
    subset = parameters["SUBSET"]
    hide = int(parameters["HIDE"]) 
    beta = int(parameters["BETA"]) 
    pred = parameters["PRED"]
    theta = float(parameters["THETA"])
    dist = int(parameters["DIST"])
    zeta = float(parameters["ZETA"])
    tscore_type = parameters["TSCORE_TYPE"]
    exp_prefix = parameters["EXP_PREFIX"]
    grouping_type = parameters["GROUPING"] 
    m_type = parameters["TYPE"]
         
    dataset_dir = os.path.abspath("../camber-pred/datasets/"+dataset)+"/"

    if beta == -1:
        beta_str = str("%.2f" % float(len(drug_res_profile.getResistant())/float(len(drug_res_profile.getSusceptible()))))
        beta = float(beta_str)

    tree.setObservedResProfile(drug_res_profile.full_profile)
    calcSubleavesResCount(tree)
    calcSubnodesCount(tree)

    newtree = removeUnResLeaves(tree)
     
    calcSubleavesResCount(newtree)
    calcSubnodesCount(newtree)
    calcDetParsResModel(newtree)
    newtree.calcContributions(drug_res_profile, beta)

    calcRefSatesMutations(all_mutations, drug_res_profile)
    calcBinSatesMutations(all_mutations, drug_res_profile)
    calcGenesCoverSupports(all_mutations, drug_res_profile, beta)

    con_mutations, neu_mutations, ess_mutations = categorizeMultiMutations(all_mutations, drug_res_profile)
    imp_mutations = con_mutations + neu_mutations + ess_mutations
    gene_profiles = geneMutations(imp_mutations)
    con_gene_profiles = geneMutations(con_mutations)
    
    gene_profiles = groupGeneMutations(gene_profiles, dist, con_gene_profiles, grouping_type)
    
    computeGeneProfiles(gene_profiles)
    computeGeneProfilesEss(gene_profiles)
    
    calcGenesCoverSupports(gene_profiles, drug_res_profile, beta)
    calcGenesWeightedSupports(gene_profiles, drug_res_profile, newtree, beta)
    calcGenesTreeScore(gene_profiles, drug_res_profile, newtree, theta=theta, zeta=zeta)
    calcOddsRatio(gene_profiles, drug_res_profile)
    
    for gene_profile in gene_profiles:
        muts_sorted = sorted(gene_profile.mutations, key=lambda mutation:mutation.tscore)
        gene_profile.start = muts_sorted[0].position
        gene_profile.end = muts_sorted[0].position
        gene_profile.desc = muts_sorted[0].desc
        
    
    if grouping_type == "G":
        gene_profiles_sorted = sorted(gene_profiles, key=lambda gp:-gp.best_mutation.wsupport)
        genes_to_show = list(gene_profiles_sorted)[:1000]        
    elif details == "N":
        genes_to_show = filerGenesToShow(gene_profiles, drug_res_profile, drug_res_genes=drug_res_genes, drug_targets=drug_targets)
    elif details == "H":
        genes_to_show = filerGenesToShowHigh(gene_profiles, max_support, max_support2=max_support2, drug_res_genes=drug_res_genes, drug_targets=drug_targets)
    elif details == "Y":
        genes_to_show = filerGenesToShowBySupport(gene_profiles, 1, drug_res_genes=drug_res_genes, target_genes=drug_targets)
    elif details == "S":
        gene_profiles_sorted = sorted(gene_profiles, key=lambda gp:-gp.wsupport)
        genes_to_show = list(gene_profiles_sorted)[:1000]

    alignments = {}

    for leaf_id in newtree.leaf_ids:
        strain_id = strains.getStrain(leaf_id)
        alignments[strain_id] = ""
    
    for i in range(0, min(10, len(genes_to_show)), 1):
        gene_profile = genes_to_show[i]
        gp_score = gene_profile.tscore
        
        for strain_index in newtree.leaf_ids:
            strain_id = strains.getStrain(strain_index)
            alignments[strain_id] += gene_profile.bin_profile[strain_index]
#        
    fasta_txt = ""
    for strain_id in alignments:
        fasta_txt += ">" + strain_id + "\n" + alignments[strain_id] + "\n"
#    
    output_fh = open(dataset_dir +"/tree_mapping/"+drug_name+"_aln"+".txt", "w")            
    output_fh.write(fasta_txt)
    output_fh.close()

    output_fh = open(dataset_dir +"/tree_mapping/"+drug_name+"_muts"+".txt", "w")
    for i in range(0, min(10, len(genes_to_show)), 1):
        gene_profile = genes_to_show[i]
        
        newtree.setObservedMutProfile(gene_profile.bin_profile)
        calcSubleavesMutCount(newtree)
        calcDetParsMutModel(newtree)
        mutated_node_leaves = calcMutatedNodeLeaves(newtree)
        line_text = gene_profile.gene_id + "_" + str(gene_profile.start)
        for mutated_node_id in mutated_node_leaves.keys():
            line_text += "\t"
            for leaf_id in mutated_node_leaves[mutated_node_id]:
                
                line_text += strains.getStrain(leaf_id) + ";"
            line_text = line_text[:-1]
            
        output_fh.write(line_text + "\n")
    output_fh.close()

    output_fh = open(dataset_dir +"/tree_mapping/"+drug_name+"_res"+".txt", "w")            
    resistant_node_leaves = calcResistantNodeLeaves(newtree)
    line_text = ""
    for resistant_node_id in resistant_node_leaves.keys():
        line_text += "\t"
        for leaf_id in resistant_node_leaves[resistant_node_id]:
            line_text += strains.getStrain(leaf_id) + ";"
        line_text = line_text[:-1]   
    output_fh.write(line_text + "\n") 
    output_fh.close()

    output_fh = open(dataset_dir +"/tree_mapping/"+drug_name+"_tree"+".txt", "w")            
    output_fh.write(newtree.toString(strains = strains, lengths = False, int_labels = True)+"\n");
    output_fh.close()


    return None


def evaluateMethod(drug_name, res_profiles, drug_targets, drug_res_genes, tree, all_mutations, gene_presence, strains, parameters):
    dataset = parameters["D"] 
    details = parameters["DETAILS"]
    subset = parameters["SUBSET"]
    hide = int(parameters["HIDE"]) 
    theta = parameters["THETA"]
    beta = int(parameters["BETA"])  
    pred = parameters["PRED"]
    zeta = float(parameters["ZETA"])
    tscore_type = parameters["TSCORE_TYPE"]
    dist = int(parameters["DIST"])
    exp_prefix = parameters["EXP_PREFIX"]
    grouping_type = parameters["GROUPING"] 
    m_type = parameters["TYPE"] 
    
    input_dir = os.path.abspath("../../../data_input/"+dataset+"_res/")+"/"
    output_dir = os.path.abspath("../../../data_output/"+dataset+"_out/")+"/"
        
    drug_res_profile = copy.deepcopy(res_profiles[drug_name])
    scoreMutations(drug_res_profile, drug_name, drug_targets[drug_name], drug_res_genes[drug_name], tree, all_mutations, gene_presence, strains, parameters=parameters)
#     output_format_fh.close()

def scoreDrug(parameters):
    dataset = parameters["D"] 
    subset = parameters["SUBSET"]
    m_type = (parameters["TYPE"])
    pred = (parameters["PRED"])
    drug_name = parameters["DRUG_NAME"]

    dataset_dir = os.path.abspath("../camber-pred/datasets/"+dataset)+"/"

    input_fh = open(dataset_dir + "strain_profiles_ord.txt")
    strains = readStrains(input_fh)
    input_fh.close()

    input_fh = open(dataset_dir + "strain_names_ord.txt")
    strain_names_ord = readStrainNamesOrd(input_fh)
    input_fh.close()

    if pred == "B":
        input_fn = dataset_dir + "res_profiles_ret.txt"
    else:
        input_fn = dataset_dir + "pred_res_profiles.txt"
        
    input_fh = open(input_fn)
    res_profiles_list = readResistanceProfiles(input_fh, strains)
    input_fh.close()
    
    res_profiles = {}
    for res_profile in res_profiles_list:
        res_profiles[res_profile.drug_name] = res_profile
    
    cluster_fh = open(dataset_dir + "cluster_names.txt")
    cluster_names, cluster_ids, cluster_ids_rev = readClusterNames(cluster_fh, strains)
    cluster_fh.close()
    
    input_fh = open(dataset_dir + "drug_targets.txt")
    drug_targets = loadDrugTargets(input_fh, cluster_ids_rev)
    input_fh.close()
    
    input_fh = open(dataset_dir + "drug_res_genes.txt")
    drug_res_genes = loadDrugResGenes(input_fh, cluster_ids_rev)
    input_fh.close()
   
    input_fh = open(dataset_dir + "drug_res_rel.txt")
    drug_res_rel = loadDrugTargets(input_fh, cluster_ids_rev)
    input_fh.close()
    
    print("DRUG RES GENES: ", drug_res_genes)
    print("DRUG TARGETS: ", drug_targets)

    if subset == "Y":
        input_fh = open(dataset_dir + 'subset_imp_genes_neigh.txt')
        gene_subset = readGeneSubset(input_fh, cluster_ids_rev)
        input_fh.close()
    else:
        gene_subset = None
        
    input_fh = open(dataset_dir + 'phylo_tree.txt')
    tree = readPhyloTree(input_fh, strains)
    input_fh.close()
        
    input_fh = open(dataset_dir + 'man-all-mutations.txt')
    all_mutations, gene_presence = readMultiMutations(input_fh, strains, cluster_names=cluster_names, cluster_ids=cluster_ids, cluster_ids_rev=cluster_ids_rev, strains_ord=strain_names_ord, gene_subset=gene_subset, m_type=m_type, syn=False)
    input_fh.close()
    
    print("READ", len(all_mutations))
    sys.stdout.flush()
    
    imp_mutations = filterMutations(all_mutations, strains)
    
    diffTime("FILTERED")
    print("MUTATIONS: ", len(imp_mutations))
    evaluateMethod(drug_name, res_profiles, drug_targets, drug_res_genes, tree, imp_mutations, gene_presence, strains, parameters)
    diffTime("EVALUATED")
    
    return drug_name
    
if __name__ == '__main__':
    
    parameters = overwriteParameters(sys.argv)
    parameters = readParameters()
    
    
    if "W" in parameters:
        WORKERS = int(parameters["W"])
    else:
        WORKERS = 4
    WORKERS = 1
    #dataset = "sta_plas_184"
    
    if "DATASET" in parameters and not parameters["D"] in ["mtb","sta","ba"]:
        dataset = parameters["D"]
    else:
        dataset = "spe"
        
    dataset = "mtb79"
    dataset_dir = os.path.abspath("../camber-pred/datasets/"+dataset)+"/"

    print(dataset_dir)
    
    ensure_dir(dataset_dir)
    ensure_dir(dataset_dir + "/html_subset/")
    ensure_dir(dataset_dir + "/r_subset/")
    ensure_dir(dataset_dir + "/tree_mapping/")
    ensure_dir(dataset_dir + "/det/")
    #scripts_dir = "/research/wongls-group/wozniak/scripts_pred/"
    #scripts_dir = "../../"+dataset+"-scripts_pred/"
    #ensure_dir(scripts_dir)
    
    cur_dir = os.path.abspath(os.curdir) +"/"
    
    hide_params = set(["0"])
    dist_params = set(["0"])
    pred_params = set(["B"])
    beta_params = set(["-1"])
    grouping_params = set(["N"])
    details_params = set(["S"])
    subset_params = set(["N"])
    type_params = set(["S"])
    theta_params = ["0.5"]
    zeta_params = ["0.0"]
    tscore_params = ["P"]
    exp_prefix = "FINAL7"
    
    input_fh = open(dataset_dir + "res_profiles_ret.txt")
    drug_names = readDrugNames(input_fh)
    input_fh.close()
    print(drug_names)
        
    TASKS = []
    
    TASK = {}
    count = 0
    for theta_param in theta_params:
        for beta_param in beta_params:
            for pred_param in pred_params:
                for type_param in type_params:
                    for hide_param in hide_params:
                        for dist_param in dist_params:
                            for tscore_param in  tscore_params:
                                for zeta_param in zeta_params:
                                    for grouping_param in grouping_params:
                                        for details_param in details_params:
                                            for subset_param in subset_params:
                                                for drug_name in drug_names:
                                                    if not drug_name in ["Ofloxacin"]:
                                                        continue 
                                                    if type_param == "G" and dist_param != "0":
                                                        continue
                                                    if grouping_param == "G" and dist_param != "0":
                                                        continue
                                                    #if drug_name == "Rifampicin":
                                                    #    continue
                                                    TASK["D"] = dataset
                                                    TASK["DETAILS"] = str(details_param)
                                                    TASK["HIDE"] = str(hide_param)
                                                    TASK["SUBSET"] = str(subset_param)
                                                    TASK["DIST"] = str(dist_param)
                                                    TASK["GROUPING"] = str(grouping_param)
                                                    TASK["EXP_PREFIX"] = exp_prefix 
                                                    TASK["TYPE"] = type_param
                                                    TASK["PRED"] = pred_param
                                                    TASK["BETA"] = beta_param
                                                    TASK["THETA"] = theta_param
                                                    TASK["TSCORE_TYPE"] = tscore_param
                                                    TASK["ZETA"] = zeta_param  
                                                    TASK["DRUG_NAME"] = drug_name
                                                    #output_det_fn = output_dir + "/det/"+str(exp_prefix)+"-"+str(0)+"-"+str(dist_param)+"-"+grouping_param+"-"+details_param+"-"+subset_param+"-"+type_param+"-"+pred_param+"-"+str(beta_param)+"-"+drug_name+".txt"
        
                                                    TASKS.append(TASK.copy())
            
                                                  #  scoreDrug(TASK)

    for TASK in TASKS:
        print(TASK)
    if WORKERS > 1:                               
        pool = Pool(processes=WORKERS)   
        for r in pool.imap(scoreDrug, TASKS):
            print(r)
    else:
        for T in TASKS:
            scoreDrug(T)



